# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from iceberg.api.types import BinaryType, TypeID


class IcebergToAvro(object):

    @staticmethod
    def type_to_schema(struct_type, name):
        struct_fields = list()
        for field in struct_type.fields:
            struct_fields.append({"field-id": field.id,
                                  "name": field.name,
                                  "type": IcebergToAvro.get_field(field)})
        return {"type": "record",
                "name": name,
                "fields": struct_fields}

    @staticmethod
    def get_field(field):
        if field.type.is_primitive_type():
            return IcebergToAvro.to_option(field)

        elif field.type.type_id == TypeID.STRUCT:
            struct_fields = list()
            for struct_field in field.type.fields:
                field_dict = {"field-id": struct_field.id,
                              "name": struct_field.name,
                              "type": IcebergToAvro.get_field(struct_field)}
                if struct_field.is_optional:
                    field_dict["default"] = None

                struct_fields.append(field_dict)

            return {"fields": struct_fields,
                    "name": field.name,
                    "type": "record"}

        elif field.type.type_id == TypeID.LIST:
            array_obj = {'element-id': field.type.element_id,
                         "items": IcebergToAvro.get_field(field.type.element_field),
                         "type": 'array'}
            if field.is_optional:
                return ['null', array_obj]
            return array_obj

        elif field.type.type_id == TypeID.MAP:
            key = field.type.key_field
            value = field.type.value_field
            array_obj = {"items": {"fields": [{"field-id": key.field_id,
                                               "name": key.name,
                                               "type": IcebergToAvro.get_field(key)},
                                              {"field-id": value.field_id,
                                               "name": value.name,
                                               "type": IcebergToAvro.get_field(value)}],
                                   "name": "k{}_v{}".format(key.field_id, value.field_id),
                                   "type": "record"},
                         "logicalType": "map",
                         "type": "array"}
            if field.is_optional:
                return ["null", array_obj]

            return array_obj

    @staticmethod
    def to_option(field):
        if field.type == BinaryType.get():
            type_name = "bytes"
        else:
            type_name = str(field.type)

        if field.is_optional:
            return ["null", type_name]
        else:
            return type_name



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from copy import deepcopy


class PackingIterator(object):

    def __init__(self, items, target_weight, lookback, weight_func):
        self.items = deepcopy(items)
        self.target_weight = target_weight
        self.lookback = lookback
        self.weight_func = weight_func
        self.bins = list()

    def __iter__(self):
        return self

    def __next__(self):
        item_copy = list(self.items)
        i = 0
        for item in item_copy:
            weight = self.weight_func(item)
            curr_bin = PackingIterator.find(self.bins, weight)
            if curr_bin is not None:
                curr_bin.add(item, weight)
                self.items.pop(i)
                i -= 1
            else:
                curr_bin = Bin(self.target_weight)
                self.items.pop(i)
                i -= 1
                curr_bin.add(item, weight)
                self.bins.append(curr_bin)

                if len(self.bins) > self.lookback:
                    return list(self.bins.pop(0).items)
            i += 1

        if len(self.bins) == 0:
            raise StopIteration()

        return list(self.bins.pop(0).items)

    @staticmethod
    def find(bins, weight):
        for curr_bin in bins:
            if curr_bin.can_add(weight):
                return curr_bin


class Bin(object):

    def __init__(self, target_weight):
        self.bin_weight = 0
        self.target_weight = target_weight
        self.items = list()

    def can_add(self, weight):
        return self.bin_weight + weight <= self.target_weight

    def add(self, item, weight):
        self.bin_weight += weight
        self.items.append(item)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import gzip
import json

from .config_properties import ConfigProperties
from .partition_spec_parser import PartitionSpecParser
from .schema_parser import SchemaParser
from .snapshot_parser import SnapshotParser
from .table_metadata import (SnapshotLogEntry,
                             TableMetadata)


class TableMetadataParser(object):
    FORMAT_VERSION = "format-version"
    LOCATION = "location"
    LAST_UPDATED_MILLIS = "last-updated-ms"
    LAST_COLUMN_ID = "last-column-id"
    SCHEMA = "schema"
    PARTITION_SPEC = "partition-spec"
    PARTITION_SPECS = "partition-specs"
    DEFAULT_SPEC_ID = "default-spec-id"
    PROPERTIES = "properties"
    CURRENT_SNAPSHOT_ID = "current-snapshot-id"
    SNAPSHOTS = "snapshots"
    SNAPSHOT_ID = "snapshot-id"
    TIMESTAMP_MS = "timestamp-ms"
    LOG = "snapshot-log"

    @staticmethod
    def to_json(metadata, indent=4):
        return json.dumps({TableMetadataParser.FORMAT_VERSION: TableMetadata.TABLE_FORMAT_VERSION,
                           TableMetadataParser.LOCATION: metadata.location,
                           TableMetadataParser.LAST_UPDATED_MILLIS: metadata.last_updated_millis,
                           TableMetadataParser.LAST_COLUMN_ID: metadata.last_column_id,
                           TableMetadataParser.SCHEMA: SchemaParser.to_dict(metadata.schema),
                           TableMetadataParser.PARTITION_SPEC: PartitionSpecParser.to_json_fields(metadata.spec),
                           TableMetadataParser.DEFAULT_SPEC_ID: int(metadata.default_spec_id),
                           TableMetadataParser.PARTITION_SPECS: [PartitionSpecParser.to_dict(spec)
                                                                 for spec in metadata.specs],
                           TableMetadataParser.PROPERTIES: metadata.properties,
                           TableMetadataParser.CURRENT_SNAPSHOT_ID: (metadata.current_snapshot_id
                                                                     if metadata.current_snapshot_id is not None
                                                                     else -1),
                           TableMetadataParser.SNAPSHOTS: [SnapshotParser.to_dict(snapshot)
                                                           for snapshot in metadata.snapshots],
                           TableMetadataParser.LOG: [{TableMetadataParser.TIMESTAMP_MS: log_entry.timestamp_millis,
                                                      TableMetadataParser.SNAPSHOT_ID: log_entry.snapshot_id}
                                                     for log_entry in metadata.snapshot_log]}, indent=indent)

    @staticmethod
    def write(metadata, metadata_location):
        if metadata_location.location().endswith(".gz"):
            output_file = gzip.open(metadata_location.create("wb"), "wb")
        else:
            output_file = metadata_location.create("wb")

        json_str = TableMetadataParser.to_json(metadata)
        output_file.write(json_str.encode("utf-8"))
        output_file.close()

    @staticmethod
    def get_file_extension(config):
        return ".metadata.json.gz" if ConfigProperties.should_compress(config) else ".metadata.json"

    @staticmethod
    def read(ops, file):
        metadata = "".join([line.decode("utf-8") for line in file.new_stream(gzipped=file.location().endswith("gz"))])
        return TableMetadataParser.from_json(ops, file.location(), metadata)

    @staticmethod
    def from_json(ops, file, json_obj):
        if isinstance(json_obj, str):
            json_obj = json.loads(json_obj)

        if not isinstance(json_obj, dict):
            raise RuntimeError("Cannot parse metadata from non-object: %s" % json_obj)

        format_version = json_obj.get(TableMetadataParser.FORMAT_VERSION)
        if format_version != TableMetadata.TABLE_FORMAT_VERSION:
            raise RuntimeError("Cannot read unsupported version: %s" % format_version)

        location = json_obj.get(TableMetadataParser.LOCATION)
        last_assigned_column = json_obj.get(TableMetadataParser.LAST_COLUMN_ID)
        schema = SchemaParser.from_json(json_obj.get(TableMetadataParser.SCHEMA))

        spec_array = json_obj.get(TableMetadataParser.PARTITION_SPECS)
        if spec_array is not None:
            default_spec_id = json_obj.get(TableMetadataParser.DEFAULT_SPEC_ID)
            specs = [PartitionSpecParser.from_json(schema, spec)
                     for spec in spec_array]
        else:
            default_spec_id = TableMetadata.INITIAL_SPEC_ID
            specs = (PartitionSpecParser.from_json_fields(schema,
                                                          default_spec_id,
                                                          json_obj.get(TableMetadataParser.PARTITION_SPEC)),)

        props = json_obj.get(TableMetadataParser.PROPERTIES)
        current_version_id = json_obj.get(TableMetadataParser.CURRENT_SNAPSHOT_ID)
        last_updated_millis = json_obj.get(TableMetadataParser.LAST_UPDATED_MILLIS)
        snapshots = [SnapshotParser.from_json(ops, snapshot) for snapshot in json_obj.get(TableMetadataParser.SNAPSHOTS)]
        entries = [SnapshotLogEntry(log_entry.get(TableMetadataParser.TIMESTAMP_MS),
                                    log_entry.get(TableMetadataParser.SNAPSHOT_ID))
                   for log_entry in sorted(json_obj.get(TableMetadataParser.LOG, []),
                                           key=lambda x: x.get(TableMetadataParser.TIMESTAMP_MS))]

        return TableMetadata(ops, file, location,
                             last_updated_millis, last_assigned_column, schema, default_spec_id, specs, props, current_version_id,
                             snapshots, entries)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import copy
import json

from iceberg.api.struct_like import StructLike


class PartitionData(StructLike):

    def __init__(self, schema=None, partition_type=None):
        if partition_type is None:
            self.partition_type = schema.as_nested_type().as_struct_type()
        else:
            for field in partition_type.fields:
                if not field.type.is_primitive_type():
                    raise RuntimeError("Partitions cannot contain nested types: %s" % field.type)
            self.partition_type = partition_type
            # schema = PartitionData.get_schema(self.partition_type)
        self._size = len(self.partition_type.fields)
        self.data = list()
        self.schema = schema
        self.string_schema = str(schema)

    def clear(self):
        self.data = list()

    def copy(self):
        return copy.deepcopy(self)

    def get_partition_type(self):
        return self.partition_type

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, PartitionData):
            return False

        return self.partition_type == other.partition_type and self.data == other.data

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return PartitionData.__class__, self.partition_type, self.data

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "PartitionData{%s}" % (",".join(["{}={}".format(self.partition_type.fields[i],
                                                               datum) for i, datum in enumerate(self.data)]))

    def __len__(self):
        return self._size

    def put(self, i, v):
        self.data.insert(i, v)

    def get(self, pos):
        try:
            return self.data[pos][1]
        except IndexError:
            return None

    @staticmethod
    def from_json(schema, json_obj):
        if isinstance(json_obj, str):
            json_obj = json.loads(json_obj)

        data = PartitionData(partition_type=schema)
        for i, kv_tuple in enumerate(json_obj.items()):
            data.put(i, kv_tuple)
        return data

    def set(self, pos, value):
        raise NotImplementedError()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from decimal import Decimal
import math
from typing import List

from .type import (Type,
                   TypeID)
from .types import (ListType,
                    MapType,
                    NestedField,
                    StructType)
from ...exceptions import ValidationException

MAX_PRECISION = list()
REQUIRED_LENGTH = [-1 for item in range(40)]

MAX_PRECISION.append(0)
for i in range(1, 24):
    MAX_PRECISION.append(int(math.floor(math.log10(math.pow(2, 8 * i - 1) - 1))))

for i in range(len(REQUIRED_LENGTH)):
    for j in range(len(MAX_PRECISION)):
        if i <= MAX_PRECISION[j]:
            REQUIRED_LENGTH[i] = j
            break
    if REQUIRED_LENGTH[i] < 0:
        raise RuntimeError("Could not find required length for precision %s" % i)


def select(schema, field_ids):
    import iceberg.api.schema
    if schema is None:
        raise RuntimeError("Schema cannot be None")
    if field_ids is None:
        raise RuntimeError("Field ids cannot be None")

    result = visit(schema, PruneColumns(field_ids))
    if schema.as_struct() == result:
        return schema
    elif result is not None:
        if schema.get_aliases() is not None:
            return iceberg.api.schema.Schema(result.as_nested_type().fields, schema.get_aliases())
        else:
            return iceberg.api.schema.Schema(result.as_nested_type().fields)

    return iceberg.api.schema.Schema(list(), schema.get_aliases())


def get_projected_ids(schema):
    import iceberg.api.schema
    if isinstance(schema, iceberg.api.schema.Schema):
        return visit(schema, GetProjectedIds())
    elif isinstance(schema, Type):
        if schema.is_primitive_type():
            return set()

        return set(visit(schema, GetProjectedIds))

    else:
        raise RuntimeError("Argument %s must be Schema or a Type" % schema)


def select_not(schema, field_ids):
    projected_ids = get_projected_ids(schema)
    projected_ids.difference(field_ids)

    return select(schema, projected_ids)


def join(left, right):
    import iceberg.api.schema
    return iceberg.api.schema.Schema(left + right)


def index_by_name(struct):
    return visit(struct, IndexByName())


def index_by_id(struct):
    return visit(struct, IndexById())


def assign_fresh_ids(type_var, next_id):
    from ..schema import Schema
    if isinstance(type_var, Type):
        return visit(type_var, AssignFreshIds(next_id))
    elif isinstance(type_var, Schema):
        schema = type_var
        return Schema(list(visit(schema.as_struct(), AssignFreshIds(next_id))
                           .as_nested_type().fields))


def decimal_to_bytes(_, value):
    scale = abs(value.as_tuple().exponent)
    quantized_value = value.quantize(Decimal("10")**-scale)
    unscaled_value = int((quantized_value * 10**scale).to_integral_value())
    min_num_bytes = (unscaled_value.bit_length() + 7) // 8
    return unscaled_value.to_bytes(min_num_bytes, 'big', signed=True)


def visit(arg, visitor): # noqa: ignore=C901
    from ..schema import Schema
    if isinstance(visitor, CustomOrderSchemaVisitor):
        return visit_custom_order(arg, visitor)
    elif isinstance(arg, Schema):
        return visitor.schema(arg, visit(arg.as_struct(), visitor))
    elif isinstance(arg, Type):
        type_var = arg
        if type_var.type_id == TypeID.STRUCT:
            struct = type_var.as_nested_type().as_struct_type()
            results = list()
            for field in struct.fields:
                visitor.field_ids.append(field.field_id)
                visitor.field_names.append(field.name)
                result = None
                try:
                    result = visit(field.type, visitor)
                except NotImplementedError:
                    # will remove it after missing functions are implemented.
                    pass
                finally:
                    visitor.field_ids.pop()
                    visitor.field_names.pop()
                results.append(visitor.field(field, result))
            return visitor.struct(struct, results)
        elif type_var.type_id == TypeID.LIST:
            list_var = type_var.as_nested_type().as_list_type()
            visitor.field_ids.append(list_var.element_id)
            try:
                element_result = visit(list_var.element_type, visitor)
            except NotImplementedError:
                # will remove it after missing functions are implemented.
                pass
            finally:
                visitor.field_ids.pop()

            return visitor.list(list_var, element_result)
        elif type_var.type_id == TypeID.MAP:
            map_var = type_var.as_nested_type().as_map_type()
            visitor.field_ids.append(map_var.key_field.field_id)
            visitor.field_names.append(map_var.key_field.name)
            try:
                key_result = visit(map_var.key_type(), visitor)
            except NotImplementedError:
                # will remove it after missing functions are implemented.
                pass
            finally:
                visitor.field_ids.pop()
                visitor.field_names.pop()

            visitor.field_ids.append(map_var.value_field.field_id)
            visitor.field_names.append(map_var.value_field.name)
            try:
                value_result = visit(map_var.value_type(), visitor)
            except NotImplementedError:
                # will remove it after missing functions are implemented.
                pass
            finally:
                visitor.field_ids.pop()
                visitor.field_names.pop()
            return visitor.map(map_var, key_result, value_result)
        else:
            return visitor.primitive(arg.as_primitive_type())
    else:
        raise RuntimeError("Invalid type for arg: %s" % arg)


def visit_custom_order(arg, visitor):
    from ..schema import Schema
    if isinstance(arg, Schema):
        schema = arg
        return visitor.schema(arg, VisitFuture(schema.as_struct(), visitor))
    elif isinstance(arg, Type):
        type_var = arg
        if type_var.type_id == TypeID.STRUCT:
            struct = type_var.as_nested_type().as_struct_type()
            results = list()
            fields = struct.fields
            for field in fields:
                results.append(VisitFieldFuture(field, visitor))
            struct = visitor.struct(struct, [x.get() for x in results])
            return struct
        elif type_var.type_id == TypeID.LIST:
            list_var = type_var.as_nested_type().as_list_type()
            return visitor.list(list_var, VisitFuture(list_var.element_type, visitor))
        elif type_var.type_id == TypeID.MAP:
            raise NotImplementedError()

        return visitor.primitive(type_var.as_primitive_type())


class SchemaVisitor(object):

    def __init__(self):
        self.field_names = list()
        self.field_ids = list()

    def schema(self, schema, struct_result):
        return None

    def struct(self, struct, field_results):
        return None

    def field(self, field, field_result):
        return None

    def list(self, list_var, element_result):
        return None

    def map(self, map_var, key_result, value_result):
        return None

    def primitive(self, primitive_var):
        return None


class CustomOrderSchemaVisitor(object):
    def __init__(self):
        super(CustomOrderSchemaVisitor, self).__init__()

    def schema(self, schema, struct_result):
        return None

    def struct(self, struct, field_results):
        return None

    def field(self, field, field_result):
        return None

    def list(self, list_var, element_result):
        return None

    def map(self, map_var, key_result, value_result):
        return None

    def primitive(self, primitive_var):
        return None


class VisitFuture(object):

    def __init__(self, type, visitor):
        self.type = type
        self.visitor = visitor

    def get(self):
        return visit(self.type, self.visitor)


class VisitFieldFuture(object):

    def __init__(self, field, visitor):
        self.field = field
        self.visitor = visitor

    def get(self):
        return self.visitor.field(self.field, VisitFuture(self.field.type, self.visitor).get)


def decimal_required_bytes(precision):
    if precision < 0 or precision > 40:
        raise RuntimeError("Unsupported decimal precision: %s" % precision)

    return REQUIRED_LENGTH[precision]


class GetProjectedIds(SchemaVisitor):

    def __init__(self):
        super(GetProjectedIds, self).__init__()
        self.field_ids = list()

    def schema(self, schema, struct_result):
        return self.field_ids

    def struct(self, struct, field_results):
        return self.field_ids

    def field(self, field, field_result):
        if field_result is None:
            self.field_ids.append(field.field_id)

        return self.field_ids

    def list(self, list_var, element_result):
        if element_result is None:
            for field in list_var.fields():
                self.field_ids.append(field.field_id)

        return self.field_ids

    def map(self, map_var, key_result, value_result):
        if value_result is None:
            for field in map_var.fields():
                self.field_ids.append(field.field_id)

        return self.field_ids


class PruneColumns(SchemaVisitor):

    def __init__(self, selected):
        super(PruneColumns, self).__init__()
        self.selected = list(selected)

    def schema(self, schema, struct_result):
        return struct_result

    def struct(self, struct, field_results):
        fields = struct.fields
        selected_fields = list()
        same_types = True

        for i, projected_type in enumerate(field_results):
            field = fields[i]
            if projected_type is not None:
                if field.type == projected_type:
                    selected_fields.append(field)
                elif projected_type is not None:
                    same_types = False
                    if field.is_optional:
                        selected_fields.append(NestedField.optional(field.field_id,
                                                                    field.name,
                                                                    projected_type))
                    else:
                        selected_fields.append(NestedField.required(field.field_id,
                                                                    field.name,
                                                                    projected_type))

        if len(selected_fields) != 0:
            if len(selected_fields) == len(fields) and same_types:
                return struct
            else:
                return StructType.of(selected_fields)

    def field(self, field, field_result):
        if field.field_id in self.selected:
            return field.type
        elif field_result is not None:
            return field_result

    def primitive(self, primitive_var):
        return None


class IndexByName(SchemaVisitor):

    DOT = "."

    def __init__(self):
        super(IndexByName, self).__init__()
        self.name_to_id = dict()

    def schema(self, schema, struct_result):
        return self.name_to_id

    def struct(self, struct, field_results):
        return self.name_to_id

    def field(self, field, field_result):
        self.add_field(field.name, field.field_id)

    def list(self, list_var, element_result):
        for field in list_var.fields():
            self.add_field(field.name, field.field_id)

    def map(self, map_var, key_result, value_result):
        for field in map_var.fields():
            self.add_field(field.name, field.field_id)

    def add_field(self, name, field_id):
        full_name = name
        if self.field_names is not None and len(self.field_names) > 0:
            full_name = IndexByName.DOT.join([IndexByName.DOT.join(self.field_names), name])

        existing_field_id = self.name_to_id.get(full_name)
        ValidationException.check(existing_field_id is None, "Invalid schema: multiple fields for name %s: %s and %s",
                                  (full_name, existing_field_id, field_id))

        self.name_to_id[full_name] = field_id


class IndexById(SchemaVisitor):

    def __init__(self):
        super(IndexById, self).__init__()
        self.index = dict()

    def schema(self, schema, struct_result):
        return self.index

    def struct(self, struct, field_results):
        return self.index

    def field(self, field, field_result):
        self.index[field.field_id] = field

    def list(self, list_var, element_result):
        for field in list_var.fields():
            self.index[field.field_id] = field

    def map(self, map_var, key_result, value_result):
        for field in map_var.fields():
            self.index[field.field_id] = field


class NextID(object):
    def __init__(self):
        raise NotImplementedError()

    def get(self):
        raise NotImplementedError()


class AssignFreshIds(CustomOrderSchemaVisitor):
    def __init__(self, next_id):
        super(AssignFreshIds, self).__init__()
        self.next_id = next_id

    def schema(self, schema, struct_result):
        return self.next_id()

    def struct(self, struct, field_results):
        fields = struct.fields
        length = len(struct.fields)
        new_ids = list()

        for _ in range(length):
            new_ids.append(self.next_id())

        new_fields = list()
        types = iter(field_results)
        for i in range(length):
            field = fields[i]
            type = next(types)
            if field.is_optional:
                new_fields.append(NestedField.optional(new_ids[i], field.name, type))
            else:
                new_fields.append(NestedField.required(new_ids[i], field.name, type))

        return StructType.of(new_fields)

    def field(self, field, field_result):
        return field_result()

    def list(self, list_var, element_result):
        new_id = self.next_id()
        if list_var.is_element_optional():
            return ListType.of_optional(new_id, element_result.get())
        else:
            return ListType.of_required(new_id, element_result.get())

    def map(self, map_var, key_result, value_result):
        new_key_id = self.next_id()
        new_value_id = self.next_id()

        if map_var.is_value_optional():
            return MapType.of_optional(new_key_id, new_value_id, key_result(), value_result())
        else:
            return MapType.of_required(new_key_id, new_value_id, key_result(), value_result())

    def primitive(self, primitive_var):
        return primitive_var


class CheckCompatibility(CustomOrderSchemaVisitor):

    @staticmethod
    def write_compatibility_errors(read_schema, write_schema):
        visit(read_schema, CheckCompatibility(write_schema, True))

    @staticmethod
    def read_compatibility_errors(read_schema, write_schema):
        visit(write_schema, CheckCompatibility(read_schema, False))

    NO_ERRORS: List[str] = []

    def __init__(self, schema, check_ordering):
        self.schema = schema
        self.check_ordering = check_ordering
        self.current_type = None

    def schema(self, schema, struct_result):
        self.current_type = self.schema.as_struct()
        try:
            struct_result.get()
        finally:
            self.current_type = None

    def struct(self, struct, field_results):
        if struct is None:
            raise RuntimeError("Evaluation must start with a schema.")

        if not self.current_type.is_struct_type():
            return [": %s cannot be read as a struct" % self.current_type]

        errors = []

        for field_errors in field_results:
            errors = errors + field_errors

        if self.check_ordering:
            new_struct = self.current_type.as_struct_type()
            id_to_ord = {}
            for i, val in enumerate(new_struct.fields):
                id_to_ord[val.field_id] = i

            last_ordinal = -1

            for read_field in self.struct.fields:
                id_var = read_field.field_id

                field = struct.field(id=id_var)
                if field is not None:
                    ordinal = id_to_ord[id]
                    if last_ordinal >= ordinal:
                        errors.append("%s is out of order before %s" % (read_field.name,
                                                                        new_struct.fields[last_ordinal].name))
                    last_ordinal = ordinal

        return errors

    def field(self, field, field_result) -> List[str]:
        struct = self.current_type.as_struct_type()
        curr_field = struct.field(field.field_id)
        errors = []

        if curr_field is None:
            if not field.is_optional:
                errors.append("{} is required, but is missing".format(field.name))
            return self.NO_ERRORS

        self.current_type = curr_field.type

        try:
            if not field.is_optional and curr_field.is_optional:
                errors.append(field.name + " should be required, but is optional")

            for error in field_result:
                if error.startswith(":"):
                    errors.append("{}{}".format(field.field_name, error))
                else:
                    errors.append("{}.{}".format(field.field_name, error))

            return errors
        except RuntimeError:
            pass
        finally:
            self.current_type = struct
        return errors

    def list(self, list_var, element_result):
        raise NotImplementedError()

    def map(self, map_var, key_result, value_result):
        raise NotImplementedError()

    def primitive(self, primitive_var):
        raise NotImplementedError()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from contextlib import contextmanager
import logging
import time

_logger = logging.getLogger(__name__)


@contextmanager
def profile(label, stats_dict=None):
    if stats_dict is None:
        _logger.debug('PROFILE: %s starting' % label)
    start = time.time()
    yield
    took = int((time.time() - start) * 1000)
    if stats_dict is None:
        _logger.debug('PROFILE: %s completed in %dms' % (label, took))
    else:
        stats_dict[label] = stats_dict.get(label, 0) + took



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import struct
import uuid


class TableOperations(object):

    def current(self):
        raise NotImplementedError()

    def refresh(self):
        raise NotImplementedError()

    def commit(self, base, metadata):
        raise NotImplementedError()

    def io(self):
        raise NotImplementedError()

    def metadata_file_location(self, file):
        raise NotImplementedError()

    def new_snapshot_id(self):
        new_uuid = uuid.uuid4()
        msb, lsb = struct.unpack(">qq", new_uuid.bytes)
        return abs(msb ^ lsb)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


import json

from iceberg.api import (ManifestFile,
                         StructLike)
from iceberg.core.avro import IcebergToAvro

from .generic_partition_field_summary import GenericPartitionFieldSummary


class GenericManifestFile(ManifestFile, StructLike):
    AVRO_SCHEMA = IcebergToAvro.type_to_schema(ManifestFile.schema().as_struct(), "manifest_file")

    GET_POS_MAP = {0: lambda curr_manifest: curr_manifest.manifest_path,
                   1: lambda curr_manifest: curr_manifest.lazy_length(),
                   2: lambda curr_manifest: curr_manifest.spec_id,
                   3: lambda curr_manifest: curr_manifest.snapshot_id,
                   4: lambda curr_manifest: curr_manifest.added_files_count,
                   5: lambda curr_manifest: curr_manifest.existing_files_count,
                   6: lambda curr_manifest: curr_manifest.deleted_files_count,
                   7: lambda curr_manifest: curr_manifest.partitions}

    @staticmethod
    def generic_manifest_from_file(file, spec_id):
        return GenericManifestFile(file=file, spec_id=spec_id)

    @staticmethod
    def generic_manifest_from_path(path, length, spec_id, snapshot_id,
                                   added_files_count, existing_files_count, deleted_files_count,
                                   partitions):
        return GenericManifestFile(path, length, spec_id, snapshot_id,
                                   added_files_count, existing_files_count, deleted_files_count,
                                   partitions)

    def __init__(self, path=None, file=None, spec_id=None, length=None, snapshot_id=None,
                 added_files_count=None, existing_files_count=None, deleted_files_count=None,
                 partitions=None):
        if file is not None:
            self.file = file
            self.manifest_path = file.location()
        else:
            self.manifest_path = path

        self._length = length
        self.spec_id = spec_id
        self.snapshot_id = snapshot_id
        self._added_files_count = added_files_count
        self._existing_files_count = existing_files_count
        self._deleted_files_count = deleted_files_count
        self.partitions = partitions
        self.from_projection_pos = None

    @property
    def length(self):
        return self.lazy_length()

    @property
    def added_files_count(self):
        return self._added_files_count

    @property
    def existing_files_count(self):
        return self._existing_files_count

    @property
    def deleted_files_count(self):
        return self._deleted_files_count

    def lazy_length(self):
        if self._length is None:
            if self.file is not None:
                self._length = self.file.get_length()
            else:
                return None
        else:
            return self._length

    def size(self):
        return len(ManifestFile.schema().columns())

    def get(self, pos, cast_type=None):
        if self.from_projection_pos:
            pos = self.from_projection_pos[pos]

        get_func = GenericManifestFile.GET_POS_MAP.get(pos)
        if get_func is None:
            raise RuntimeError("Unknown field ordinal: %s" % pos)

        if cast_type is not None:
            return cast_type(get_func(self))

        return get_func(self)

    def set(self, pos, value):
        if self.from_projection_pos:
            pos = self.from_projection_pos[pos]

        if pos == 0:
            self.manifest_path = str(value)
        elif pos == 1:
            self._length = int(value)
        elif pos == 2:
            self.spec_id = int(value)
        elif pos == 3:
            self.snapshot_id = int(value)
        elif pos == 4:
            self._added_files_count = int(value)
        elif pos == 5:
            self._existing_files_count = int(value)
        elif pos == 6:
            self._deleted_files_count = int(value)
        elif pos == 7:
            self.partitions = value

    def copy(self):
        return GenericManifestFile(path=self.manifest_path, spec_id=self.spec_id, length=self.length,
                                   snapshot_id=self.snapshot_id, added_files_count=self.added_files_count,
                                   existing_files_count=self.existing_files_count,
                                   deleted_files_count=self.deleted_files_count,
                                   partitions=list(self.partitions))

    @staticmethod
    def get_schema():
        return GenericManifestFile.AVRO_SCHEMA

    @staticmethod
    def to_avro_record_json(manifest):
        return json.dumps(GenericManifestFile.to_avro_record_dict(manifest))

    @staticmethod
    def to_avro_record_dict(manifest):
        return {"manifest_path": manifest.manifest_path,
                "manifest_length": manifest._length,
                "partition_spec_id": manifest.spec_id,
                "added_snapshot_id": manifest.snapshot_id,
                "added_data_files_count": manifest.added_files_count,
                "existing_data_files_count": manifest.existing_files_count,
                "deleted_data_files_count": manifest.deleted_files_count,
                "partitions": manifest.partitions}

    @staticmethod
    def from_avro_record_json(row):
        partitions = row.get("partitions")
        if partitions is not None:
            partitions = [GenericPartitionFieldSummary(contains_null=partition["contains_null"],
                                                       lower_bound=partition["lower_bound"],
                                                       upper_bound=partition["upper_bound"])
                          for partition in row.get("partitions")]
        return GenericManifestFile(path=row.get("manifest_path"),
                                   length=row.get("manifest_length"),
                                   spec_id=row.get("partition_spec_id"),
                                   added_files_count=row.get("added_data_files_count"),
                                   existing_files_count=row.get("existing_data_files_count"),
                                   deleted_files_count=row.get("existing_data_files_count"),
                                   partitions=partitions)

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, GenericManifestFile):
            return False

        return self.__key() == other.__key()

    def __hash__(self):
        return hash(self.__key())

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "GenericManifestFile({})".format(self.manifest_path)

    def __key(self):
        return (GenericManifestFile.__class__,
                self.manifest_path)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from iceberg.api import Table

from .data_table_scan import DataTableScan
from .schema_update import SchemaUpdate


class BaseTable(Table):

    def __init__(self, ops, name):
        self.ops = ops
        self.name = name

    def refresh(self):
        self.ops.refresh()

    def new_scan(self):
        return DataTableScan(self.ops, self)

    def schema(self):
        return self.ops.current().schema

    def spec(self):
        return self.ops.current().spec

    def properties(self):
        return self.ops.current().properties

    def location(self):
        return self.ops.current().location

    def current_snapshot(self):
        return self.ops.current().current_snapshot()

    def snapshots(self):
        return self.ops.current().snapshots

    def snapshots_with_summary_property(self, prop_key, prop_val):
        if prop_key is None:
            raise RuntimeError("Property Key cannot be None: (%s, %s)" % (prop_key, prop_val))

        for snapshot in self.ops.current().snapshots:
            if prop_key in snapshot.summary.keys() and snapshot.summary.get(prop_key) == prop_val:
                yield snapshot

    def update_schema(self):
        return SchemaUpdate(self.ops)

    def update_properties(self):
        # PropertiesUpdate(self.ops)
        raise NotImplementedError()

    def update_location(self):
        # SetLocation(self.ops)
        raise NotImplementedError()

    def new_append(self):
        # MergeAppend(ops)
        raise NotImplementedError()

    def new_fast_append(self):
        # FastAppend(ops)
        raise NotImplementedError()

    def new_rewrite(self):
        # ReplaceFiles(ops)
        raise NotImplementedError()

    def new_overwrite(self):
        # OverwriteData(ops)
        raise NotImplementedError()

    def new_replace_partitions(self):
        # ReplacePartitionsOperation(ops)
        raise NotImplementedError()

    def new_delete(self):
        # StreamingDelete(ops)
        raise NotImplementedError()

    def expire_snapshots(self):
        # RemoveSnapshots(ops)
        raise NotImplementedError()

    def rollback(self):
        # RollbackToSnapshot(ops)
        raise NotImplementedError()

    def new_transaction(self):
        # BaseTransaction.newTransaction(ops)
        raise NotImplementedError()

    def __str__(self):
        return self.name



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import gzip
import os

from iceberg.exceptions import AlreadyExistsException

from .expressions import JAVA_MAX_INT
from .io import (InputFile,
                 OutputFile,
                 PositionOutputStream,
                 SeekableInputStream)


class Files(object):

    @staticmethod
    def local_output(file):
        return LocalOutputFile(file)

    @staticmethod
    def local_input(file):
        return LocalInputFile(file)


class LocalOutputFile(OutputFile):

    def __init__(self, file):
        self.file = file

    def create(self):
        if os.path.exists(self.file):
            raise AlreadyExistsException("File already exists: %s" % self.file)

        return PositionOutputStream(open(self.file, "w"))

    def create_or_overwrite(self):
        if os.path.exists(self.file):
            if not os.remove(self.file.name):
                raise RuntimeError("File deleting file: %s" % self.file)

        return self.create()

    def location(self):
        return self.file

    def __str__(self):
        return self.location()


class LocalInputFile(InputFile):

    def __init__(self, file):
        self.file = file

    def get_length(self):
        return os.path.getsize(self.file)

    def new_stream(self, gzipped=False):
        with open(self.file, "rb") as fo:
            if gzipped:
                fo = gzip.GzipFile(fileobj=fo)
            return fo

    def location(self):
        return self.file

    def __str__(self):
        return self.location()


class SeekableFileInputStream(SeekableInputStream):

    def __init__(self, stream):
        self.stream = stream

    def get_pos(self):
        return self.stream.tell()

    def seek(self, new_pos):
        return self.stream.seek(new_pos)

    def read(self, b=None, off=None, read_len=None):
        if b is None and off is None and read_len is None:
            return None, self.stream.read()
        if b is None:
            raise RuntimeError("supplied byte field is None")

        if off is None and read_len is None:
            new_b = self.stream.read(b.length)
            return len(new_b), new_b

        if off is not None and read_len is None or off is None and read_len is not None:
            raise RuntimeError("Invalid args: read_len(%s), off(%s)" % (read_len, off))

        if read_len < 0 or off < 0 or (len(b) - off < read_len):
            raise RuntimeError("Invalid args: read_len(%s), off(%s), len_b_offset(%s)" % (read_len, off, len(b) - off))

        new_b = bytes(self.stream.read(read_len), "utf8")

        if off > 0:
            new_b = b[0:off] + new_b
        if off + read_len < len(b):
            new_b = new_b + b[off + read_len:]
        return read_len, new_b

    def skip(self, n):
        if n > JAVA_MAX_INT:
            return self.stream.seek(JAVA_MAX_INT)
        else:
            return self.stream.seek(n)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from datetime import datetime
import logging

from iceberg.api import Filterable
from iceberg.api import TableScan
from iceberg.api.expressions import (Binder,
                                     Expressions)
from iceberg.api.io import CloseableGroup
from iceberg.api.types import get_projected_ids, select

from .base_combined_scan_task import BaseCombinedScanTask
from .table_properties import TableProperties
from .util import PackingIterator

_logger = logging.getLogger(__name__)


class BaseTableScan(CloseableGroup, TableScan):
    DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
    SNAPSHOT_COLUMNS = ("snapshot_id", "file_path", "file_ordinal", "file_format", "block_size_in_bytes",
                        "file_size_in_bytes", "record_count", "partition", "value_counts", "null_value_counts",
                        "lower_bounds", "upper_bounds")

    def new_refined_scan(self, ops, table, schema, snapshot_id, row_filter,
                         case_sensitive, selected_columns, options, minused_cols):
        raise NotImplementedError()

    def target_split_size(self, ops):
        raise NotImplementedError()

    def __init__(self, ops, table, schema, snapshot_id=None, columns=None,
                 row_filter=None, case_sensitive=True, selected_columns=None, options=None,
                 minused_cols=None):
        self.ops = ops
        self.table = table
        self._schema = schema
        self.snapshot_id = snapshot_id
        self.columns = columns
        self._row_filter = row_filter
        self._case_sensitive = case_sensitive
        self.selected_columns = selected_columns
        self.minused_cols = minused_cols or list()
        self.options = options if options is not None else dict()

        if self.columns is None and self._row_filter is None:
            self.columns = Filterable.ALL_COLUMNS
            self._row_filter = Expressions.always_true()

        self._stats = dict()

    def is_case_sensitive(self):
        return self.case_sensitive

    def use_snapshot(self, snapshot_id):
        if self.snapshot_id is not None:
            raise RuntimeError("Cannot override snapshot, already set to id=%s" % self.snapshot_id)
        if self.ops.current().snapshot(snapshot_id) is None:
            raise RuntimeError("Cannot find snapshot with ID %s" % self.snapshot_id)

        return self.new_refined_scan(self.ops, self.table, self._schema, snapshot_id=snapshot_id,
                                     row_filter=self._row_filter, case_sensitive=self._case_sensitive,
                                     selected_columns=self.selected_columns, options=self.options,
                                     minused_cols=self.minused_cols)

    def as_of_time(self, timestamp_millis):
        raise NotImplementedError()

    def project(self, schema):
        return self.new_refined_scan(self.ops, self.table, schema, snapshot_id=self.snapshot_id,
                                     row_filter=self._row_filter, case_sensitive=self._case_sensitive,
                                     selected_columns=self.selected_columns, options=self.options,
                                     minused_cols=self.minused_cols)

    def case_sensitive(self, case_sensitive):
        return self.new_refined_scan(self.ops, self.table, self._schema, snapshot_id=self.snapshot_id,
                                     row_filter=self._row_filter, case_sensitive=case_sensitive,
                                     selected_columns=self.selected_columns, options=self.options,
                                     minused_cols=self.minused_cols)

    def select(self, columns):
        return self.new_refined_scan(self.ops, self.table, self._schema, snapshot_id=self.snapshot_id,
                                     row_filter=self._row_filter, case_sensitive=self._case_sensitive,
                                     selected_columns=columns, options=self.options,
                                     minused_cols=self.minused_cols)

    def select_except(self, columns):
        return self.new_refined_scan(self.ops, self.table, self._schema, snapshot_id=self.snapshot_id,
                                     row_filter=self._row_filter, case_sensitive=self._case_sensitive,
                                     selected_columns=self.selected_columns, options=self.options,
                                     minused_cols=columns)

    @property
    def row_filter(self):
        return self._row_filter

    def filter(self, expr):
        return self.new_refined_scan(self.ops, self.table, self._schema, snapshot_id=self.snapshot_id,
                                     row_filter=Expressions.and_(self._row_filter, expr),
                                     case_sensitive=self._case_sensitive, selected_columns=self.selected_columns,
                                     options=self.options, minused_cols=self.minused_cols)

    def option(self, property, value):
        builder = dict()
        builder.update(self.options)
        builder[property] = value
        return self.new_refined_scan(self.ops, self.table, self._schema, snapshot_id=self.snapshot_id,
                                     row_filter=self._row_filter, case_sensitive=self._case_sensitive,
                                     selected_columns=self.selected_columns, options=builder,
                                     minused_cols=self.minused_cols)

    def plan_files(self, ops=None, snapshot=None, row_filter=None):

        if not all(i is None for i in [ops, snapshot, row_filter]):
            raise NotImplementedError()

        snapshot = self.ops.current().snapshot(self.snapshot_id) \
            if self.snapshot_id is not None else self.ops.current().current_snapshot()

        if snapshot is not None:
            _logger.info("Scanning table {} snapshot {} created at {} with filter {}"
                         .format(self.table,
                                 snapshot.snapshot_id,
                                 datetime.fromtimestamp(snapshot.timestamp_millis / 1000.0)
                                 .strftime(BaseTableScan.DATE_FORMAT),
                                 self._row_filter))

            return self.plan_files(ops, snapshot, row_filter)
        else:
            _logger.info("Scanning empty table {}" % self.table)

    def plan_tasks(self):
        split_size = self.target_split_size(self.ops)
        lookback = int(self.ops.current().properties.get(TableProperties.SPLIT_LOOKBACK,
                                                         TableProperties.SPLIT_LOOKBACK_DEFAULT))
        open_file_cost = int(self.ops.current().properties.get(TableProperties.SPLIT_OPEN_FILE_COST,
                                                               TableProperties.SPLIT_OPEN_FILE_COST_DEFAULT))

        if not self.ops.conf.get("iceberg.scan.split-file-tasks", True):
            split_files = list(self.plan_files())
        else:
            split_files = self.split_files(split_size)

        def weight_func(file):
            return max(file.length, open_file_cost)

        return (BaseCombinedScanTask(scan_tasks)
                for scan_tasks in PackingIterator(split_files, split_size, lookback, weight_func))

    def split_files(self, split_size):
        file_scan_tasks = list(self.plan_files())
        split_tasks = [task for split_tasks in [scan_task.split(split_size) for scan_task in file_scan_tasks]
                       for task in split_tasks]

        return split_tasks

    @property
    def schema(self):
        return self._lazy_column_projection()

    def to_arrow_table(self):
        raise NotImplementedError()

    def to_pandas(self):
        raise NotImplementedError()

    def _lazy_column_projection(self):
        if "*" in self.selected_columns:
            if len(self.minused_cols) == 0:
                return self._schema
            self.selected_columns = [field.name for field in self._schema.as_struct().fields]
            final_selected_cols = [column for column in self.selected_columns if column not in self.minused_cols]
        else:
            final_selected_cols = self.selected_columns

        required_field_ids = set()
        required_field_ids.update(Binder.bound_references(self.table.schema().as_struct(),
                                                          [self._row_filter],
                                                          self._case_sensitive))

        if self._case_sensitive:
            selected_ids = get_projected_ids(self.table.schema().select(final_selected_cols))
        else:
            selected_ids = get_projected_ids(self.table.schema().case_insensitive_select(final_selected_cols))

        required_field_ids.update(selected_ids)

        return select(self.table.schema(), required_field_ids)

    def __repr__(self):
        return "BaseTableScan(table={}, projection={}, filter={}, case_sensitive={}".format(self.table,
                                                                                            self._schema.as_struct(),
                                                                                            self._row_filter,
                                                                                            self._case_sensitive)

    def __str__(self):
        return self.__repr__()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .filesystem_table_operations import FilesystemTableOperations
from .. import TableOperations
from ..table_metadata import TableMetadata
from ...api import PartitionSpec, Schema, Table, Tables
from ...exceptions import NoSuchTableException


class FilesystemTables(Tables):

    def __init__(self: "FilesystemTables", conf: dict = None) -> None:
        self.conf = conf if conf is not None else dict()

    def load(self: "FilesystemTables", table_identifier: str) -> Table:
        from ..base_table import BaseTable
        ops = self.new_table_ops(table_identifier)
        if ops.current() is None:
            raise NoSuchTableException("Table does not exist at location: %s" % table_identifier)

        return BaseTable(ops, table_identifier)

    def create(self: "FilesystemTables", schema: Schema, table_identifier: str, spec: PartitionSpec = None,
               properties: dict = None, location: str = None) -> Table:
        """
        Create a new table on the filesystem.

        Note: it is expected that the filesystem has atomic operations to ensure consistency for metadata updates.
        Filesystems that don't have this guarantee could lead to data loss.

        Location should always be None as the table location on disk is taken from `table_identifier`
        """
        from ..base_table import BaseTable
        if location:
            raise RuntimeError("""location has to be None. Both table_identifier and location have been declared.
             table_identifier: {} and location: {}""".format(table_identifier, location))

        full_spec, properties = super(FilesystemTables, self).default_args(spec, properties)
        ops = self.new_table_ops(table_identifier)

        metadata = TableMetadata.new_table_metadata(ops, schema, full_spec, table_identifier, properties)
        ops.commit(None, metadata)

        return BaseTable(ops, table_identifier)

    def new_table_ops(self: "FilesystemTables", table_identifier: str) -> TableOperations:
        if table_identifier is None:
            raise RuntimeError("table_identifier cannot be None")

        return FilesystemTableOperations(table_identifier, self.conf)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from setuptools import setup

package_metadata: dict = {}
with open("./iceberg/__init__.py") as fp:
    exec(fp.read(), package_metadata)

setup(
    name='acryl-iceberg-legacy',
    version=package_metadata["__version__"],
    maintainer='Acryl Data',
    author_email='datahub@acryl.io',
    description='Acryl maintained copy of Iceberg Python bindings. Iceberg is a new table format for storing large, slow-moving tabular data',
    keywords='iceberg',
    url='https://github.com/acryldata/py-iceberg',
    python_requires='>=3.7',
    install_requires=['botocore',
                      'boto3',
                      'fastavro>=1.3.2<1.4.0',
                      'hmsclient==0.1.1',
                      'mmh3',
                      'pyparsing>=2.4.7<2.5.0',
                      'python-dateutil',
                      'pytz',
                      'requests',
                      'retrying',
                      'pandas',
                      'pyarrow>=6.0.1',
                      'azure-storage-file-datalake',
                      'azure-identity'
                      ],
    extras_require={
        "dev": [
            "tox-travis==0.12",
            "virtualenv<20.0.0",
        ],
    },
    setup_requires=['setupmeta'],
    license="Apache License 2.0",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from decimal import Decimal, ROUND_HALF_UP
import struct
import sys
import uuid

from .type import TypeID
from .type_util import decimal_to_bytes


class Conversions(object):
    HIVE_NULL = "__HIVE_DEFAULT_PARTITION__"
    value_mapping = {TypeID.BOOLEAN: lambda as_str: as_str.lower() == "true" if as_str is not None else False,
                     TypeID.INTEGER: lambda as_str: int(float(as_str)),
                     TypeID.LONG: lambda as_str: int(float(as_str)),
                     TypeID.FLOAT: lambda as_str: float(as_str),
                     TypeID.DOUBLE: lambda as_str: float(as_str),
                     TypeID.STRING: lambda as_str: as_str,
                     TypeID.UUID: lambda as_str: uuid.UUID(as_str),
                     TypeID.FIXED: lambda as_str: bytearray(bytes(as_str, "UTF-8")
                                                            if sys.version_info >= (3, 0)
                                                            else bytes(as_str)),
                     TypeID.BINARY: lambda as_str: bytes(as_str, "UTF-8") if sys.version_info >= (3, 0) else bytes(as_str),
                     TypeID.DECIMAL: lambda as_str: Decimal(as_str),
                     }

    to_byte_buff_mapping = {TypeID.BOOLEAN: lambda type_id, value: struct.pack("<?", 1 if value else 0),
                            TypeID.INTEGER: lambda type_id, value: struct.pack("<i", value),
                            TypeID.DATE: lambda type_id, value: struct.pack("<i", value),
                            TypeID.LONG: lambda type_id, value: struct.pack("<q", value),
                            TypeID.TIME: lambda type_id, value: struct.pack("<q", value),
                            TypeID.TIMESTAMP: lambda type_id, value: struct.pack("<q", value),
                            TypeID.FLOAT: lambda type_id, value: struct.pack("<f", value),
                            TypeID.DOUBLE: lambda type_id, value: struct.pack("<d", value),
                            TypeID.STRING: lambda type_id, value: value.encode('UTF-8'),
                            TypeID.UUID: lambda type_id, value: struct.pack('>QQ', (value.int >> 64)
                                                                            & 0xFFFFFFFFFFFFFFFF, value.int
                                                                            & 0xFFFFFFFFFFFFFFFF),
                            TypeID.FIXED: lambda type_id, value: value,
                            TypeID.BINARY: lambda type_id, value: value,
                            TypeID.DECIMAL: decimal_to_bytes
                            }

    from_byte_buff_mapping = {TypeID.BOOLEAN: lambda type_var, value: struct.unpack('<?', value)[0] != 0,
                              TypeID.INTEGER: lambda type_var, value: struct.unpack('<i', value)[0],
                              TypeID.DATE: lambda type_var, value: struct.unpack('<i', value)[0],
                              TypeID.LONG: lambda type_var, value: struct.unpack('<q', value)[0],
                              TypeID.TIME: lambda type_var, value: struct.unpack('<q', value)[0],
                              TypeID.TIMESTAMP: lambda type_var, value: struct.unpack('<q', value)[0],
                              TypeID.FLOAT: lambda type_var, value: struct.unpack('<f', value)[0],
                              TypeID.DOUBLE: lambda type_var, value: struct.unpack('<d', value)[0],
                              TypeID.STRING: lambda type_var, value: bytes(value).decode("utf-8"),
                              TypeID.UUID: lambda type_var, value:
                              uuid.UUID(int=struct.unpack('>QQ', value)[0] << 64 | struct.unpack('>QQ', value)[1]),
                              TypeID.FIXED: lambda type_var, value: value,
                              TypeID.BINARY: lambda type_var, value: value,
                              TypeID.DECIMAL: lambda type_var, value:
                              Decimal(int.from_bytes(value, 'big', signed=True) * 10**-type_var.scale)
                              .quantize(Decimal("." + "".join(["0" for i in range(1, type_var.scale)]) + "1"),
                                        rounding=ROUND_HALF_UP)
                              }

    @staticmethod
    def from_partition_string(type_var, as_string):
        if as_string is None or Conversions.HIVE_NULL == as_string:
            return None
        part_func = Conversions.value_mapping.get(type_var.type_id)
        if part_func is None:
            raise RuntimeError(f"Unsupported type for from_partition_string: {type_var}")

        return part_func(as_string)

    @staticmethod
    def to_byte_buffer(type_id, value):
        try:
            return Conversions.to_byte_buff_mapping[type_id](type_id, value)
        except KeyError:
            raise TypeError(f"Cannot serialize type, no conversion mapping found for TypeID: {type_id}")

    @staticmethod
    def from_byte_buffer(type_var, buffer_var):
        try:
            return Conversions.from_byte_buff_mapping[type_var.type_id](type_var, buffer_var)
        except KeyError:
            raise TypeError(f"Cannot deserialize type, no conversion mapping found for TypeID: {type_var.type_id}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from .expressions import Expressions, ExpressionVisitors, RewriteNot
from .predicate import BoundPredicate, UnboundPredicate


def inclusive(spec, case_sensitive=True):
    return InclusiveProjection(spec, case_sensitive)


def strict(spec):
    return StrictProjection(spec)


class ProjectionEvaluator(ExpressionVisitors.ExpressionVisitor):

    def project(self, expr):
        raise NotImplementedError()


class BaseProjectionEvaluator(ProjectionEvaluator):

    def __init__(self, spec, case_sensitive=True):
        self.spec = spec
        self.case_sensitive = case_sensitive

    def project(self, expr):
        # projections assume that there are no NOT nodes in the expression tree. to ensure that this
        # is the case, the expression is rewritten to push all NOT nodes down to the expression
        # leaf nodes.
        # this is necessary to ensure that the default expression returned when a predicate can't be
        # projected is correct.
        #
        return ExpressionVisitors.visit(ExpressionVisitors.visit(expr, RewriteNot.get()), self)

    def always_true(self):
        return Expressions.always_true()

    def always_false(self):
        return Expressions.always_false()

    def not_(self, result):
        raise RuntimeError("[BUG] project called on expression with a not")

    def and_(self, left_result, right_result):
        return Expressions.and_(left_result, right_result)

    def or_(self, left_result, right_result):
        return Expressions.or_(left_result, right_result)

    def predicate(self, pred):
        bound = pred.bind(self.spec.schema.as_struct(), case_sensitive=self.case_sensitive)

        if isinstance(bound, BoundPredicate):
            return self.predicate(bound)

        return bound


class InclusiveProjection(BaseProjectionEvaluator):

    def __init__(self, spec, case_sensitive=True):
        super(InclusiveProjection, self).__init__(spec,
                                                  case_sensitive=case_sensitive)

    def predicate(self, pred):
        if isinstance(pred, UnboundPredicate):
            return super(InclusiveProjection, self).predicate(pred)

        part = self.spec.get_field_by_source_id(pred.ref.field.field_id)

        if part is None:
            return self.always_true()

        result = part.transform.project(part.name, pred)
        if result is not None:
            return result

        return self.always_true()


class StrictProjection(BaseProjectionEvaluator):

    def __init__(self, spec):
        super(StrictProjection, self).__init__(spec)

    def predicate(self, pred):
        part = self.spec.get_field_by_source_id(pred.ref.field.field_id)

        if part is None:
            return self.always_false()

        result = part.transform.project_strict(part.name, pred)

        if result is not None:
            return result

        return self.always_false()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from iceberg.api import (FileFormat,
                         Metrics)
from iceberg.api.types import Conversions

from .generic_data_file import GenericDataFile
from .partition_data import PartitionData


class DataFiles(object):
    DEFAULT_BLOCK_SIZE = 64 * 1024 * 1024

    @staticmethod
    def new_partition_data(spec):
        return PartitionData(spec.partition_type())

    @staticmethod
    def copy_partition_data(spec, partition_data, reuse):
        data = reuse
        if data is None:
            data = DataFiles.new_partition_data(spec)

        for i, _ in enumerate(spec.fields):
            data.set(i, partition_data.get(i))

        return data

    @staticmethod
    def fill_from_path(spec, partition_path, reuse):
        data = reuse
        if data is None:
            data = DataFiles.new_partition_data(spec)

        partitions = partition_path.split("/")
        if len(partitions) > len(spec.fields):
            raise RuntimeError("Invalid partition data, too many fields (expecting %s): %s" % (len(spec.fields),
                                                                                               partition_path))

        if len(partitions) < len(spec.fields):
            raise RuntimeError("Invalid partition data, not enough fields(expecting %s): %s" % (len(spec.fields),
                                                                                                partition_path))

        for i, part in enumerate(partitions):
            field = spec.fields[i]
            parts = part.split("=")
            if len(parts) != 2 or parts[0] is None or parts[0] != field.name:
                raise RuntimeError("Invalid partition: %s" % part)

            data.set(i, Conversions.from_partition_string(data.get_type(i), parts[1]))

        return data

    @staticmethod
    def data(spec, partition_path):
        return DataFiles.fill_from_path(spec, partition_path, None)

    @staticmethod
    def copy(spec, partition):
        return DataFiles.copy_partition_data(spec, partition, None)

    @staticmethod
    def from_input_file(input_file, row_count, partition_data=None, metrics=None):
        from .filesystem import FileSystemInputFile
        if isinstance(input_file, FileSystemInputFile):
            return DataFiles.from_stat(input_file.get_stat(), row_count,
                                       partition_data=partition_data, metrics=metrics)

    @staticmethod
    def from_stat(stat, row_count, partition_data=None, metrics=None):
        location = stat.path
        format = FileFormat.from_file_name(location)
        return GenericDataFile(location, format, stat.length, stat.block_size,
                               row_count=row_count, partition=partition_data, metrics=metrics)

    @staticmethod
    def builder(spec=None):
        return DataFileBuilder(spec)


class DataFileBuilder(object):

    def __init__(self, spec=None):
        self.spec = spec
        self.partition_data = DataFiles.new_partition_data(spec) if spec is not None else None
        self.is_partitioned = spec is not None
        self.file_path = None
        self.format = None
        self.record_count = -1
        self.file_size_in_bytes = -1
        self.block_size_in_bytes = -1
        self.column_sizes = None
        self.value_counts = None
        self.null_value_counts = None
        self.lower_bounds = None
        self.upper_bounds = None

    def clear(self):
        if self.is_partitioned:
            self.partition_data.clear()
        self.file_path = None
        self.format = None
        self.record_count = -1
        self.file_size_in_bytes = -1
        self.block_size_in_bytes = -1
        self.column_sizes = None
        self.value_counts = None
        self.null_value_counts = None
        self.lower_bounds = None
        self.upper_bounds = None
        return self

    def copy(self, to_copy):
        if self.is_partitioned:
            self.partition_data = DataFiles.copy_partition_data(self.spec, to_copy.partition, self.partition_data)

        self.file_path = str(to_copy.path())
        self.format = to_copy.format
        self.record_count = to_copy.record_count
        self.file_size_in_bytes = to_copy.file_size_in_bytes
        self.block_size_in_bytes = to_copy.block_size_in_bytes
        self.column_sizes = to_copy.column_sizes
        self.value_counts = to_copy.value_counts
        self.null_value_counts = to_copy.null_value_counts
        self.lower_bounds = to_copy.lower_bounds
        self.upper_bounds = to_copy.upper_bounds
        return self

    def with_status(self, stat):
        self.file_path = stat.path
        self.file_size_in_bytes = stat.length
        self.block_size_in_bytes = stat.blocksize
        return self

    def with_input_file(self, input_file):
        from .filesystem import FileSystemInputFile
        if isinstance(input_file, FileSystemInputFile):
            self.with_status(input_file.get_stat())

        self.file_path = self.location()
        self.file_size_in_bytes = self.get_length()

        return self

    def with_path(self, path):
        self.file_path = path
        return self

    def with_format(self, fmt):
        if isinstance(fmt, FileFormat):
            self.format = fmt
        else:
            self.format = FileFormat[str(fmt).upper()]

        return self

    def with_partition(self, partition):
        self.partition_data = DataFiles.copy_partition_data(self.spec, partition, self.partition_data)
        return self

    def with_record_count(self, record_count):
        self.record_count = record_count
        return self

    def with_file_size_in_bytes(self, file_size_in_bytes):
        self.file_size_in_bytes = file_size_in_bytes
        return self

    def with_block_size_in_bytes(self, block_size_in_bytes):
        self.block_size_in_bytes = block_size_in_bytes
        return self

    def with_partition_path(self, partition_path):
        if not self.is_partitioned:
            raise RuntimeError("Cannot add partition data for an unpartitioned table")

        self.partition_data = DataFiles.fill_from_path(self.spec, partition_path, self.partition_data)
        return self

    def with_metrics(self, metrics):
        self.record_count = metrics.row_count if metrics.row_count is not None else -1
        self.column_sizes = metrics.column_sizes
        self.value_counts = metrics.value_counts
        self.null_value_counts = metrics.null_value_counts
        self.lower_bounds = metrics.lower_bounds
        self.upper_bounds = metrics.upper_bounds
        return self

    def build(self):
        if self.file_path is None:
            raise RuntimeError("File path is required")
        if self.format is None:
            self.format = FileFormat.from_file_name(self.file_path)

        if self.format is None:
            raise RuntimeError("File format is required")

        if self.file_size_in_bytes < 0:
            raise RuntimeError("File size is required")

        if self.record_count < 0:
            raise RuntimeError("Record count is required")

        if self.block_size_in_bytes is None:
            self.block_size_in_bytes = DataFiles.DEFAULT_BLOCK_SIZE

        return GenericDataFile(self.file_path, self.format, self.file_size_in_bytes, self.block_size_in_bytes,
                               partition=self.partition_data.copy() if self.is_partitioned else None,
                               metrics=Metrics(row_count=self.record_count, column_sizes=self.column_sizes,
                                               value_counts=self.value_counts,
                                               null_value_counts=self.null_value_counts,
                                               lower_bounds=self.lower_bounds, upper_bounds=self.upper_bounds))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from decimal import Decimal
from enum import Enum, unique
import uuid


@unique
class TypeID(Enum):
    BOOLEAN = {"java_class": "Boolean.class", "python_class": bool, "id": 1}
    INTEGER = {"java_class": "Integer.class", "python_class": int, "id": 2}
    LONG = {"java_class": "Long.class", "python_class": int, "id": 3}
    FLOAT = {"java_class": "Float.class", "python_class": float, "id": 4}
    DOUBLE = {"java_class": "Double.class", "python_class": float, "id": 5}
    DATE = {"java_class": "Integer.class", "python_class": int, "id": 6}
    TIME = {"java_class": "Long.class", "python_class": int, "id": 7}
    TIMESTAMP = {"java_class": "Long.class", "python_class": int, "id": 8}
    STRING = {"java_class": "CharSequence.class", "python_class": str, "id": 9}
    UUID = {"java_class": "java.util.UUID.class", "python_class": uuid.UUID, "id": 10}
    FIXED = {"java_class": "ByteBuffer.class", "python_class": bytes, "id": 11}
    BINARY = {"java_class": "ByteBuffer.class", "python_class": bytearray, "id": 12}
    DECIMAL = {"java_class": "BigDecimal.class", "python_class": Decimal, "id": 13}
    STRUCT = {"java_class": "Void.class", "python_class": None, "id": 14}
    LIST = {"java_class": "Void.class", "python_class": None, "id": 15}
    MAP = {"java_class": "Void.class", "python_class": None, "id": 16}


class Type(object):
    length: int
    scale: int
    precision: int

    def __init__(self):
        pass

    def type_id(self):
        pass

    def is_primitive_type(self):
        return False

    def as_primitive_type(self):
        raise ValueError("Not a primitive type: " + self)

    def as_struct_type(self):
        raise ValueError("Not a struct type: " + self)

    def as_list_type(self):
        raise ValueError("Not a list type: " + self)

    def as_map_type(self):
        raise ValueError("Not a map type: " + self)

    def is_nested_type(self):
        return False

    def is_struct_type(self):
        return False

    def is_list_type(self):
        return False

    def is_map_type(self):
        return False

    def as_nested_type(self):
        raise ValueError("Not a nested type: " + self)


class PrimitiveType(Type):

    def __eq__(self, other):
        return type(self) == type(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_primitive_type(self):
        return True

    def as_primitive_type(self):
        return self


class NestedType(Type):

    def __init__(self):
        super(NestedType, self).__init__()

    def is_nested_type(self):
        return True

    def as_nested_type(self):
        return self

    def fields(self):
        pass

    def field_type(self, name):
        pass

    def field(self, id):
        pass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from urllib.parse import quote_plus

from iceberg.exceptions import ValidationException

from .partition_field import PartitionField
from .schema import Schema
from .transforms import Transform, Transforms
from .types import (NestedField,
                    StructType)


class PartitionSpec(object):

    PARTITION_DATA_ID_START = 1000

    @staticmethod
    def UNPARTITIONED_SPEC():
        return PartitionSpec(Schema(), 0, [], PartitionSpec.PARTITION_DATA_ID_START - 1)

    @staticmethod
    def unpartitioned():
        return PartitionSpec.UNPARTITIONED_SPEC()

    def __init__(self, schema, spec_id, fields, last_assigned_field_id):
        self.fields_by_source_id = None
        self.fields_by_name = None
        self.__java_classes = None
        self.field_list = None

        self.schema = schema
        self.spec_id = spec_id
        self.__fields = list()
        for field in fields:
            self.__fields.append(field)
        self.last_assigned_field_id = last_assigned_field_id

    @property
    def fields(self):
        return self.lazy_field_list()

    @property
    def java_classes(self):
        if self.__java_classes is None:
            self.__java_classes
        for field in self.__fields:
            source_type = self.schema.find_type(field.source_id)
            result = field.transform().get_result_by_type(source_type)
            self.__java_classes.append(result.type_id.java_class())

        return self.__java_classes

    def get_field_by_source_id(self, field_id):
        return self.lazy_fields_by_source_id().get(field_id)

    def partition_type(self):
        struct_fields = list()
        for _i, field in enumerate(self.__fields):
            source_type = self.schema.find_type(field.source_id)
            result_type = field.transform.get_result_type(source_type)
            struct_fields.append(NestedField.optional(field.field_id,
                                                      field.name,
                                                      result_type))

        return StructType.of(struct_fields)

    def get(self, data, pos, java_class):
        data.get(pos, java_class)

    def escape(self, string):
        return quote_plus(string, encoding="UTF-8")

    def partition_to_path(self, data):
        sb = list()
        java_classes = self.java_classes
        for i, jclass in enumerate(java_classes):
            field = self.__fields[i]
            value_string = field.transform().to_human_string(self.get(data, i, jclass))

            if i > 0:
                sb.append("/")
            sb.append(field.name)
            sb.append("=")
            sb.append(self.escape(value_string))

        return "".join(sb)

    def compatible_with(self, other):
        if self.__eq__(other):
            return True

        if len(self.__fields) != len(other.__fields):
            return False

        for i, field in enumerate(self.__fields):
            that_field = other.__fields[i]
            if field.source_id != that_field.source_id or str(field.transform) != str(that_field.transform):
                return False

        return True

    def lazy_fields_by_source_id(self):
        if self.fields_by_source_id is None:
            self.fields_by_source_id = dict()
            for field in self.fields:
                self.fields_by_source_id[field.source_id] = field

        return self.fields_by_source_id

    def identity_source_ids(self):
        source_ids = set()
        fields = self.fields
        for field in fields:
            if "identity" == str(field.transform()):
                source_ids.add(field)

        return source_ids

    def lazy_field_list(self):
        if self.field_list is None:
            self.field_list = list(self.__fields)

        return self.field_list

    def lazy_fields_by_source_name(self):
        if self.fields_by_name is None:
            self.fields_by_name = dict()
            for field in self.__fields:
                self.fields_by_name[field.name] = field

        return self.fields_by_name

    def __eq__(self, other):
        if id(self) == id(other):
            return True

        if other is None or not isinstance(other, PartitionSpec):
            return False

        return self.__fields == other.__fields

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return PartitionSpec.__class__, tuple(self.fields)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        sb = ["["]

        for field in self.__fields:
            sb.append("\n {field_id}: {name}: {transform}({source_id})".format(field_id=field.field_id,
                                                                               name=field.name,
                                                                               transform=str(field.transform),
                                                                               source_id=field.source_id))

        if len(self.__fields) > 0:
            sb.append("\n")
        sb.append("]")

        return "".join(sb)

    @staticmethod
    def builder_for(schema: Schema) -> "PartitionSpecBuilder":
        return PartitionSpecBuilder(schema)

    @staticmethod
    def check_compatibility(spec, schema):
        for field in spec.fields:
            src_type = schema.find_type(field.source_id)
            if not src_type.is_primitive_type():
                raise ValidationException("Cannot partition by non-primitive source field: %s", src_type)
            if not field.transform.can_transform(src_type):
                raise ValidationException("Invalid source type %s for transform: %s", (src_type, field.transform))


class PartitionSpecBuilder(object):

    def __init__(self, schema):
        self.schema = schema
        self.fields = list()
        self.partition_names = set()
        self.dedup_fields = dict()
        self.spec_id = 0
        self.last_assigned_field_id = PartitionSpec.PARTITION_DATA_ID_START - 1

    def __next_field_id(self):
        self.last_assigned_field_id = self.last_assigned_field_id + 1
        return self.last_assigned_field_id

    def with_spec_id(self, spec_id):
        self.spec_id = spec_id
        return self

    def check_and_add_partition_name(self, name, source_column_id=None):
        schema_field = self.schema.find_field(name)
        if source_column_id is not None:
            if schema_field is not None and schema_field.field_id != source_column_id:
                raise ValueError("Cannot create identity partition sourced from different field in schema: %s" % name)
        else:
            if schema_field is not None:
                raise ValueError("Cannot create partition from name that exists in schema: %s" % name)

        if name is None or name == "":
            raise ValueError("Cannot use empty or null partition name: %s" % name)
        if name in self.partition_names:
            raise ValueError("Cannot use partition names more than once: %s" % name)

        self.partition_names.add(name)
        return self

    def check_redundant_and_add_field(self, field_id: int, name: str, transform: Transform) -> None:
        field = PartitionField(field_id,
                               self.__next_field_id(),
                               name,
                               transform)
        dedup_key = (field.source_id, field.transform.dedup_name())
        partition_field = self.dedup_fields.get(dedup_key)
        if partition_field is not None:
            raise ValueError("Cannot add redundant partition: %s conflicts with %s" % (partition_field, field))
        self.dedup_fields[dedup_key] = field
        self.fields.append(field)

    def find_source_column(self, source_name):
        source_column = self.schema.find_field(source_name)
        if source_column is None:
            raise RuntimeError("Cannot find source column: %s" % source_name)

        return source_column

    def identity(self, source_name, target_name=None):
        if target_name is None:
            target_name = source_name

        source_column = self.find_source_column(source_name)
        self.check_and_add_partition_name(target_name, source_column.field_id)
        self.check_redundant_and_add_field(source_column.field_id,
                                           target_name,
                                           Transforms.identity(source_column.type))
        return self

    def year(self, source_name, target_name=None):
        if target_name is None:
            target_name = "{}_year".format(source_name)

        self.check_and_add_partition_name(target_name)
        source_column = self.find_source_column(source_name)
        self.check_redundant_and_add_field(source_column.field_id,
                                           target_name,
                                           Transforms.year(source_column.type))
        return self

    def month(self, source_name, target_name=None):
        if target_name is None:
            target_name = "{}_month".format(source_name)

        self.check_and_add_partition_name(target_name)
        source_column = self.find_source_column(source_name)
        self.check_redundant_and_add_field(source_column.field_id,
                                           target_name,
                                           Transforms.month(source_column.type))
        return self

    def day(self, source_name, target_name=None):
        if target_name is None:
            target_name = "{}_day".format(source_name)

        self.check_and_add_partition_name(target_name)
        source_column = self.find_source_column(source_name)
        self.check_redundant_and_add_field(source_column.field_id,
                                           target_name,
                                           Transforms.day(source_column.type))
        return self

    def hour(self, source_name, target_name=None):
        if target_name is None:
            target_name = "{}_hour".format(source_name)

        self.check_and_add_partition_name(target_name)
        source_column = self.find_source_column(source_name)
        self.check_redundant_and_add_field(source_column.field_id,
                                           target_name,
                                           Transforms.hour(source_column.type))
        return self

    def bucket(self, source_name, num_buckets, target_name=None):
        if target_name is None:
            target_name = "{}_bucket".format(source_name)

        self.check_and_add_partition_name(target_name)
        source_column = self.find_source_column(source_name)
        self.fields.append(PartitionField(source_column.field_id,
                                          self.__next_field_id(),
                                          target_name,
                                          Transforms.bucket(source_column.type, num_buckets)))
        return self

    def truncate(self, source_name, width, target_name=None):
        if target_name is None:
            target_name = "{}_truncate".format(source_name)

        self.check_and_add_partition_name(target_name)
        source_column = self.find_source_column(source_name)
        self.fields.append(PartitionField(source_column.field_id,
                                          self.__next_field_id(),
                                          target_name,
                                          Transforms.truncate(source_column.type, width)))
        return self

    def add_without_field_id(self, source_id, name, transform):
        return self.add(source_id, self.__next_field_id(), name, transform)

    def add(self, source_id: int, field_id: int, name: str, transform: str) -> "PartitionSpecBuilder":
        column = self.schema.find_field(source_id)
        if column is None:
            raise ValueError("Cannot find source column: %s" % source_id)

        transform_obj = Transforms.from_string(column.type, transform)
        self.check_and_add_partition_name(name, source_id)
        self.fields.append(PartitionField(source_id,
                                          field_id,
                                          name,
                                          transform_obj))
        self.last_assigned_field_id = max(self.last_assigned_field_id, field_id)
        return self

    def build(self):
        spec = PartitionSpec(self.schema, self.spec_id, self.fields, self.last_assigned_field_id)
        PartitionSpec.check_compatibility(spec, self.schema)
        return spec



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging

import fastavro
from iceberg.api import FileFormat, Filterable
from iceberg.api.expressions import Expressions, inclusive
from iceberg.api.io import CloseableGroup

from .avro import AvroToIceberg
from .filtered_manifest import FilteredManifest
from .manifest_entry import ManifestEntry, Status
from .partition_spec_parser import PartitionSpecParser
from .schema_parser import SchemaParser
from .table_metadata import TableMetadata

_logger = logging.getLogger(__name__)


class ManifestReader(CloseableGroup, Filterable):
    ALL_COLUMNS = ("*",)
    CHANGE_COLUMNS = ("file_path", "file_format", "partition", "record_count", "file_size_in_bytes")

    @staticmethod
    def read(file, spec_lookup=None):
        return ManifestReader(file=file, spec_lookup=spec_lookup)

    def select(self, columns):
        return FilteredManifest(self,
                                Expressions.always_true(),
                                Expressions.always_true(),
                                list(columns),
                                self.case_sensitive)

    def filter_partitions(self, expr):
        return FilteredManifest(self,
                                expr,
                                Expressions.always_true(),
                                ManifestReader.ALL_COLUMNS,
                                self.case_sensitive)

    def filter_rows(self, expr):
        return FilteredManifest(self,
                                inclusive(self.spec).project(expr),
                                expr,
                                ManifestReader.ALL_COLUMNS,
                                self.case_sensitive)

    @staticmethod
    def in_memory(spec, entries):
        return ManifestReader(spec=spec, entries=entries)

    def __init__(self, file=None, spec=None, metadata=None, schema=None, case_sensitive=True, spec_lookup=None):
        self.file = file
        self.schema = schema
        self.metadata = metadata
        self.spec = spec
        self._case_sensitive = case_sensitive

        self._entries = None
        self._avro_rows = list()
        self._fo = None
        self._avro_reader = None
        self._avro_rows = None

        if not all([item is not None for item in [self.file, self.metadata, self.spec, self.schema]]):
            if self.spec is not None:
                self.__init_from_spec()
            else:
                self.__init_from_file(spec_lookup)

        self._adds = None
        self._deletes = None

    def __init_from_file(self, spec_lookup):
        self._fo = self.file.new_fo()
        self._avro_reader = fastavro.reader(self._fo)
        self.metadata = self._avro_reader.metadata
        spec_id = int(self.metadata.get("partition-spec-id", TableMetadata.INITIAL_SPEC_ID))

        if spec_lookup is not None:
            self.spec = spec_lookup(spec_id)
            self.schema = self.spec.schema
        else:
            self.schema = SchemaParser.from_json(self.metadata.get("schema"))
            self.spec = PartitionSpecParser.from_json_fields(self.schema, spec_id, self.metadata.get("partition-spec"))

    def __init_from_spec(self):
        self.metadata = dict()
        self.schema = self.spec.schema

    def case_sensitive(self, case_sensitive):
        return ManifestReader(file=self.file, metadata=self.metadata, spec=self.spec,
                              schema=self.schema, case_sensitive=case_sensitive)

    def cache_changes(self):
        adds = list()
        deletes = list()
        for entry in self.entries(ManifestReader.CHANGE_COLUMNS):
            if entry.status == "ADDED":
                adds.append(entry.copy())
            elif entry.status == "DELETED":
                deletes.append(entry.copy())

        self._adds = adds
        self._deletes = deletes

    def added_files(self):
        if self._adds is None:
            self.cache_changes()

        return self._adds

    def deleted_files(self):
        if self._adds is None:
            self.cache_changes()

        return self._deletes

    def entries(self, columns=None):
        if columns is None:
            columns = ManifestReader.ALL_COLUMNS

        file_format = FileFormat.from_file_name(self.file.location())
        if file_format is None:
            raise RuntimeError("Unable to determine format of manifest: %s" % self.file)

        proj_schema = ManifestEntry.project_schema(self.spec.partition_type(), columns)

        if self._entries is None:
            if file_format is FileFormat.AVRO:
                self._entries = list()
                for read_entry in AvroToIceberg.read_avro_row(proj_schema, self._avro_reader):
                    entry = ManifestEntry(schema=proj_schema, partition_type=self.spec.partition_type())
                    for i, key in enumerate(read_entry.keys()):
                        entry.put(i, read_entry[key])
                    self._entries.append(entry)
                self._fo.close()
                self._avro_reader = None

        return self._entries

    def iterator(self, part_filter=None, columns=None):
        if part_filter is None and columns is None:
            return self.iterator(Expressions.always_true(), Filterable.ALL_COLUMNS)

        return [entry.file for entry in self.entries() if entry.status != Status.DELETED]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import time

from iceberg.api import (Filterable,
                         FilteredSnapshot,
                         ManifestFile,
                         Snapshot,
                         SnapshotIterable)
from iceberg.api.expressions import Expressions
from iceberg.api.io import CloseableGroup
from iceberg.core.avro import AvroToIceberg

from .generic_manifest_file import GenericManifestFile
from .manifest_reader import ManifestReader


class BaseSnapshot(Snapshot, SnapshotIterable, CloseableGroup):

    @staticmethod
    def snapshot_from_files(ops, snapshot_id, files):
        return BaseSnapshot(ops, snapshot_id, None,
                            manifests=[GenericManifestFile(file=ops.new_input_file(path), spec_id=0)
                                       for path in files])

    def __init__(self, ops, snapshot_id, parent_id=None, manifests=None, manifest_list=None, timestamp_millis=None,
                 operation=None, summary=None):
        super(BaseSnapshot, self).__init__()
        if timestamp_millis is None:
            timestamp_millis = int(time.time() * 1000)

        self._ops = ops
        self._snapshot_id = snapshot_id
        self._parent_id = parent_id
        self._timestamp_millis = timestamp_millis
        if manifests is not None:
            self._manifests = [manifest if isinstance(manifest, GenericManifestFile)
                               else GenericManifestFile(file=ops.new_input_file(manifest), spec_id=0)
                               for manifest in manifests]
        else:
            self._manifests = None
        self._manifest_list = manifest_list
        self._operation = operation
        self._summary = summary

        self._adds = None
        self._deletes = None

    @property
    def snapshot_id(self):
        return self._snapshot_id

    @property
    def timestamp_millis(self):
        return self._timestamp_millis

    @property
    def parent_id(self):
        return self._parent_id

    @property
    def manifests(self):
        if self._manifests is None:
            # if manifest isn't set then the snapshot_file is set and should be read to get the list
            return (GenericManifestFile.from_avro_record_json(manifest)
                    for manifest in AvroToIceberg.read_avro_file(ManifestFile.schema(), self._manifest_list))

        return self._manifests

    @property
    def manifest_location(self):
        return self._manifest_list.location() if self._manifest_list is not None else None

    @property
    def summary(self):
        return self._summary

    @property
    def operation(self):
        return self._operation

    def select(self, columns):
        return FilteredSnapshot(self, Expressions.always_true(), Expressions.always_true(), columns)

    def filter_partitions(self, expr):
        return FilteredSnapshot(self, expr, Expressions.always_true(), Filterable.ALL_COLUMNS)

    def filter_rows(self, expr):
        return FilteredSnapshot(self, Expressions.always_true(), expr, Filterable.ALL_COLUMNS)

    def iterator(self, part_filter=None, row_filter=None, columns=None):
        if part_filter is None and row_filter is None and columns is None:
            return self.iterator(Expressions.always_true(), Expressions.always_true(), Filterable.ALL_COLUMNS)
        return iter([self.get_filtered_manifest(path, part_filter, row_filter, columns)
                     for path in self._manifest_files])

    def added_files(self):
        raise NotImplementedError()

    def deleted_files(self):
        raise NotImplementedError()

    def cache_changes(self):
        raise NotImplementedError

    def __repr__(self):
        return "BaseSnapshot(id={id},timestamp_ms={ts_ms},manifests={manifests}".format(id=self._snapshot_id,
                                                                                        ts_ms=self._timestamp_millis,
                                                                                        manifests=self._manifests)

    def __str__(self):
        return self.__repr__()

    def get_filtered_manifest(self, path, part_filter=None, row_filter=None, columns=None):
        reader = ManifestReader.read(self._ops.new_input_file(path))
        self.add_closeable(reader)
        return reader



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from enum import Enum, unique


@unique
class FileFormat(Enum):
    ORC = {"extension": "orc", "splittable": True}
    PARQUET = {"extension": "parquet", "splittable": True}
    AVRO = {"extension": "avro", "splittable": True}

    def add_extension(self, filename):
        if filename.endswith(self.value["extension"]):
            return filename
        else:
            return filename + "." + self.value["extension"]

    def is_splittable(self):
        return self.value["splittable"]

    @staticmethod
    def from_file_name(filename):
        last_index_of = filename.rfind('.')
        if last_index_of < 0:
            return None
        ext = filename[last_index_of + 1:]
        for fmt in FileFormat:
            if ext == fmt.value["extension"]:
                return fmt



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from .expressions import Expressions, ExpressionVisitors
from .predicate import BoundPredicate, Predicate, UnboundPredicate


class ResidualEvaluator(object):

    def __init__(self, spec, expr):
        self._spec = spec
        self._expr = expr
        self.__visitor = None

    def _visitor(self):
        if self.__visitor is None:
            self.__visitor = ResidualVisitor()

        return self.__visitor

    def residual_for(self, partition_data):
        return self._visitor().eval(partition_data)


class ResidualVisitor(ExpressionVisitors.BoundExpressionVisitor):

    def __init__(self):
        self.struct = None

    def eval(self, struct):
        self.struct = struct

    def always_true(self):
        return Expressions.always_true()

    def always_false(self):
        return Expressions.always_false()

    def is_null(self, ref):
        return self.always_true() if ref.get(self.struct) is None else self.always_false()

    def not_null(self, ref):
        return self.always_true() if ref.get(self.struct) is not None else self.always_false()

    def lt(self, ref, lit):
        return self.always_true() if ref.get(self.struct) < lit.value else self.always_false()

    def lt_eq(self, ref, lit):
        return self.always_true() if ref.get(self.struct) <= lit.value else self.always_false()

    def gt(self, ref, lit):
        return self.always_true() if ref.get(self.struct) > lit.value else self.always_false()

    def gt_eq(self, ref, lit):
        return self.always_true() if ref.get(self.struct) >= lit.value else self.always_false()

    def eq(self, ref, lit):
        return self.always_true() if ref.get(self.struct) == lit.value else self.always_false()

    def not_eq(self, ref, lit):
        return self.always_true() if ref.get(self.struct) != lit.value else self.always_false()

    def not_(self, result):
        return Expressions.not_(result)

    def and_(self, left_result, right_result):
        return Expressions.and_(left_result, right_result)

    def or_(self, left_result, right_result):
        return Expressions.or_(left_result, right_result)

    def predicate(self, pred):
        if isinstance(pred, BoundPredicate):
            return self.bound_predicate(pred)
        elif isinstance(pred, UnboundPredicate):
            return self.unbound_predicate(pred)

        raise RuntimeError("Invalid predicate argument %s" % pred)

    def bound_predicate(self, pred):
        part = self.spec.get_field_by_source_id(pred.ref.field_id)
        if part is None:
            return pred

        strict_projection = part.transform.project_strict(part.name, pred)
        if strict_projection is None:
            bound = strict_projection.bind(self.spec.partition_type())
            if isinstance(bound, BoundPredicate):
                return super(ResidualVisitor, self).predicate(bound)
            return bound

        return pred

    def unbound_predicate(self, pred):
        bound = pred.bind(self.spec.schema.as_struct())

        if isinstance(bound, BoundPredicate):
            bound_residual = self.predicate(bound)
            if isinstance(bound_residual, Predicate):
                return pred
            return bound_residual

        return bound



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from iceberg.api.expressions import Expression, Operation, Predicate
import pyarrow.dataset as ds


def get_dataset_filter(expr: Expression, expected_to_file_map: dict) -> ds.Expression:
    """
    Given an Iceberg Expression and a mapping of names in the iceberg schema to the file schema,
    convert to an equivalent dataset filter using the file column names. Recursively iterate through the
    expressions to convert each portion one predicate at a time

    Parameters
    ----------
    expr : iceberg.api.expressions.Expression
        An Iceberg Expression to be converted
    expected_to_file_map : dict
        A dict that maps the iceberg schema names to the names from the file schema
    Returns
    -------
    pyarrow._dataset.Expression
        An equivalent dataset expression
    """
    if expr is None:
        return None

    if isinstance(expr, Predicate):
        return predicate(expr, expected_to_file_map)

    if expr.op() == Operation.TRUE:
        return None
    elif expr.op() == Operation.FALSE:
        return False
    elif expr.op() == Operation.NOT:
        return not_(get_dataset_filter(expr.child, expected_to_file_map))
    elif expr.op() == Operation.AND:
        return and_(get_dataset_filter(expr.left, expected_to_file_map),
                    get_dataset_filter(expr.right, expected_to_file_map))
    elif expr.op() == Operation.OR:
        return or_(get_dataset_filter(expr.left, expected_to_file_map),
                   get_dataset_filter(expr.right, expected_to_file_map))
    else:
        raise RuntimeError("Unknown operation: {}".format(expr.op()))


def predicate(pred: Predicate, field_map: dict) -> ds.Expression:  # noqa: ignore=C901
    """
    Given an Iceberg Predicate and a mapping of names in the iceberg schema to the file schema,
    convert to an equivalent dataset expression using the file column names.

    Parameters
    ----------
    pred : iceberg.api.expressions.Predicate
        An Iceberg Predicate to be converted
    field_map : dict
        A dict that maps the iceberg schema names to the names from the file schema
    Returns
    -------
    pyarrow._dataset.Expression
        An equivalent dataset expression
    """
    # get column name in the file schema so we can apply the predicate
    col_name = field_map.get(pred.ref.name)

    if col_name is None:
        if pred.op == Operation.IS_NULL:
            return ds.scalar(True) == ds.scalar(True)

        return ds.scalar(True) == ds.scalar(False)

    if pred.op == Operation.IS_NULL:
        return ~ds.field(col_name).is_valid()
    elif pred.op == Operation.NOT_NULL:
        return ds.field(col_name).is_valid()
    elif pred.op == Operation.LT:
        return ds.field(col_name) < pred.lit.value
    elif pred.op == Operation.LT_EQ:
        return ds.field(col_name) <= pred.lit.value
    elif pred.op == Operation.GT:
        return ds.field(col_name) > pred.lit.value
    elif pred.op == Operation.GT_EQ:
        return ds.field(col_name) >= pred.lit.value
    elif pred.op == Operation.EQ:
        return ds.field(col_name) == pred.lit.value
    elif pred.op == Operation.NOT_EQ:
        return ds.field(col_name) != pred.lit.value
    elif pred.op == Operation.IN:
        return ds.field(col_name).isin(pred.lit.value)
    elif pred.op == Operation.NOT_IN:
        return ds.field(col_name).isin(pred.lit.value)


def and_(left: ds.Expression, right: ds.Expression) -> ds.Expression:
    """
    Given a left and right expression combined them using the `AND` logical operator

    Parameters
    ----------
    left : pyarrow._dataset.Expression
        A Dataset `Expression` to logically `AND`
    right : pyarrow._dataset.Expression
       A Dataset `Expression` to logically `AND`
    Returns
    -------
    pyarrow._dataset.Expression
        The left and right `Expression` combined with `AND`
    """
    return left & right


def or_(left: ds.Expression, right: ds.Expression) -> ds.Expression:
    """
    Given a left and right expression combined them using the `OR` logical operator

    Parameters
    ----------
    left : pyarrow._dataset.Expression
        A Dataset `Expression` to logically `OR`
    right : pyarrow._dataset.Expression
       A Dataset `Expression` to logically `OR`
    Returns
    -------
    pyarrow._dataset.Expression
        The left and right `Expression` combined with `OR`
    """
    return left | right


def not_(child: ds.Expression) -> ds.Expression:
    """
    Given a child expression create the logical negation

    Parameters
    ----------
    child : pyarrow._dataset.Expression
        A Dataset `Expression` to logically `OR`
    Returns
    -------
    pyarrow._dataset.Expression
        The negation of the input `Expression`
    """
    return ~child



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from iceberg.api import FileScanTask

from .partition_spec_parser import PartitionSpecParser
from .schema_parser import SchemaParser


class BaseFileScanTask(FileScanTask):

    def __init__(self, file, schema_str, spec_str, residuals):
        self._file = file
        self._schema_str = schema_str
        self._spec_str = spec_str
        self._spec = None
        self._residuals = residuals

    @property
    def file(self):
        return self._file

    @property
    def spec(self):
        if self._spec is None:
            self._spec = PartitionSpecParser.from_json(SchemaParser.from_json(self._schema_str), self._spec_str)

        return self._spec

    @property
    def start(self):
        return 0

    @property
    def length(self):
        return self._file.file_size_in_bytes()

    @property
    def residual(self):
        return self._residuals.residual_for(self._file.partition())

    def split(self, split_size):
        if self.file.format().is_splittable():
            return [task for task in SplitScanTaskIterator(split_size, self)]
        else:
            return self

    def __repr__(self):
        fields = ["file: {}".format(self._file.path()),
                  "partition_data: {}".format(self._file.partition()),
                  "residual: {}".format(self.residual)]

        return "BaseFileScanTask({})".format(", ".join(fields))

    def __str__(self):
        return self.__repr__()


class SplitScanTaskIterator(object):

    def __init__(self, split_size, file_scan_task):
        self._offset = 0
        self._remaining_len = file_scan_task.length
        self._split_size = split_size
        self._file_scan_task = file_scan_task

    def has_next(self):
        return self._remaining_len > 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.has_next():
            len = min(self._split_size, self._remaining_len)
            split_task = SplitScanTask(self._offset, len, self._file_scan_task)
            self._offset += len
            self._remaining_len -= len
            return split_task

        raise StopIteration


class SplitScanTask(FileScanTask):

    def __init__(self, offset, len, file_scan_task):
        self._offset = offset
        self._len = len
        self._file_scan_task = file_scan_task

    @property
    def file(self):
        return self._file_scan_task.file

    @property
    def spec(self):
        return self._file_scan_task.spec

    @property
    def start(self):
        return self._offset

    @property
    def length(self):
        return self._len

    @property
    def residual(self):
        return self._file_scan_task.residual()

    def split(self):
        raise RuntimeError("Cannot split a task which is already split")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from math import isnan
from typing import Any, List, Optional, TYPE_CHECKING, Union

from iceberg.exceptions import ValidationException

from .expression import (Expression,
                         Operation)
from .literals import (BaseLiteral,
                       Literals)
from .term import BoundTerm, UnboundTerm
from ..types import TypeID

if TYPE_CHECKING:
    from iceberg.api import StructLike


class Predicate(Expression):

    def __init__(self, op: Operation, term: Union[BoundTerm, UnboundTerm]):
        if term is None:
            raise ValueError("Term cannot be None")

        self.op: Operation = op
        self.term: Union[BoundTerm, UnboundTerm] = term

    @property
    def ref(self):
        return self.term.ref

    @property
    def lit(self):
        raise NotImplementedError("Not Implemented for base class")

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        elif other is None or not isinstance(other, Predicate):
            return False

        return self.op == other.op and self.ref == other.ref and self.lit == other.lit

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "Predicate({},{},{})".format(self.op, self.ref, self.lit)

    def __str__(self):
        if self.op == Operation.IS_NULL:
            return "is_null({})".format(self.ref)
        elif self.op == Operation.NOT_NULL:
            return "not_null({})".format(self.ref)
        elif self.op == Operation.LT:
            return "less_than({})".format(self.ref)
        elif self.op == Operation.LT_EQ:
            return "less_than_equal({})".format(self.ref)
        elif self.op == Operation.GT:
            return "greater_than({})".format(self.ref)
        elif self.op == Operation.GT_EQ:
            return "greater_than_equal({})".format(self.ref)
        elif self.op == Operation.EQ:
            return "equal({})".format(self.ref)
        elif self.op == Operation.NOT_EQ:
            return "not_equal({})".format(self.ref)
        else:
            return "invalid predicate: operation = {}".format(self.op)


class BoundPredicate(Predicate):

    def __init__(self, op: Operation, term: BoundTerm, lit: BaseLiteral = None, literals: List[BaseLiteral] = None,
                 is_unary_predicate: bool = False, is_literal_predicate: bool = False,
                 is_set_predicate: bool = False):
        self.is_unary_predicate = is_unary_predicate
        self.is_literal_predicate = is_literal_predicate
        self.is_set_predicate = is_set_predicate

        super(BoundPredicate, self).__init__(op, term)
        ValidationException.check(sum([is_unary_predicate, is_literal_predicate, is_set_predicate]) == 1,
                                  "Only a single predicate type may be set: %s=%s, %s=%s, %s=%s",
                                  ("is_unary_predicate", is_unary_predicate,
                                   "is_literal_predicate", is_literal_predicate,
                                   "is_set_predicate", is_set_predicate))

        self._literals: Optional[List[BaseLiteral]] = None
        if self.is_unary_predicate:
            ValidationException.check(lit is None, "Unary Predicates may not have a literal", ())

        elif self.is_literal_predicate:
            ValidationException.check(lit is not None, "Literal Predicates must have a literal set", ())
            self._literals = [lit]  # type: ignore

        elif self.is_set_predicate:
            ValidationException.check(literals is not None, "Set Predicates must have literals set", ())
            self._literals = literals
        else:
            raise ValueError(f"Unable to instantiate {op} -> (lit={lit}, literal={literals}")

    @property
    def lit(self) -> Optional[BaseLiteral]:
        if self._literals is None or len(self._literals) == 0:
            return None
        return self._literals[0]

    def eval(self, struct: StructLike) -> bool:
        ValidationException.check(isinstance(self.term, BoundTerm), "Term must be bound to eval: %s", (self.term))
        return self.test(self.term.eval(struct))  # type: ignore

    def test(self, struct: StructLike = None, value: Any = None) -> bool:
        ValidationException.check(struct is None or value is None, "Either struct or value must be none", ())
        if struct is not None:
            ValidationException.check(isinstance(self.term, BoundTerm), "Term must be bound to eval: %s", (self.term))
            return self.test(value=self.term.eval(struct))  # type: ignore
        else:
            if self.is_unary_predicate:
                return self.test_unary_predicate(value)
            elif self.is_literal_predicate:
                return self.test_literal_predicate(value)
            else:
                return self.test_set_predicate(value)

    def test_unary_predicate(self, value: Any) -> bool:

        if self.op == Operation.IS_NULL:
            return value is None
        elif self.op == Operation.NOT_NULL:
            return value is not None
        elif self.op == Operation.IS_NAN:
            return isnan(value)
        elif self.op == Operation.NOT_NAN:
            return not isnan(value)
        else:
            raise ValueError(f"{self.op} is not a valid unary predicate")

    def test_literal_predicate(self, value: Any) -> bool:
        if self.lit is None:
            raise ValidationException("Literal must not be none", ())

        if self.op == Operation.LT:
            return value < self.lit.value
        elif self.op == Operation.LT_EQ:
            return value <= self.lit.value
        elif self.op == Operation.GT:
            return value > self.lit.value
        elif self.op == Operation.GT_EQ:
            return value >= self.lit.value
        elif self.op == Operation.EQ:
            return value == self.lit.value
        elif self.op == Operation.NOT_EQ:
            return value != self.lit.value
        else:
            raise ValueError(f"{self.op} is not a valid literal predicate")

    def test_set_predicate(self, value: Any) -> bool:
        if self._literals is None:
            raise ValidationException("Literals must not be none", ())

        if self.op == Operation.IN:
            return value in self._literals
        elif self.op == Operation.NOT_IN:
            return value not in self._literals
        else:
            raise ValueError(f"{self.op} is not a valid set predicate")


class UnboundPredicate(Predicate):

    def __init__(self, op, term, value=None, lit=None, values=None, literals=None):
        self._literals = None
        num_set_args = sum([1 for x in [value, lit, values, literals] if x is not None])

        if num_set_args > 1:
            raise ValueError(f"Only one of value={value}, lit={lit}, values={values}, literals={literals} may be set")
        super(UnboundPredicate, self).__init__(op, term)
        if isinstance(value, BaseLiteral):
            lit = value
            value = None
        if value is not None:
            self._literals = [Literals.from_(value)]
        elif lit is not None:
            self._literals = [lit]
        elif values is not None:
            self._literals = map(Literals.from_, values)
        elif literals is not None:
            self._literals = literals

    @property
    def literals(self):
        return self._literals

    @property
    def lit(self):
        if self.op in [Operation.IN, Operation.NOT_IN]:
            raise ValueError(f"{self.op} predicate cannot return a literal")

        return None if self.literals is None else self.literals[0]

    def negate(self):
        return UnboundPredicate(self.op.negate(), self.term, literals=self.literals)

    def bind(self, struct, case_sensitive=True):
        bound = self.term.bind(struct, case_sensitive=case_sensitive)

        if self.literals is None:
            return self.bind_unary_operation(bound)
        elif self.op in [Operation.IN, Operation.NOT_IN]:
            return self.bind_in_operation(bound)

        return self.bind_literal_operation(bound)

    def bind_unary_operation(self, bound_term: BoundTerm) -> BoundPredicate:
        from .expressions import Expressions
        if self.op == Operation.IS_NULL:
            if bound_term.ref.field.is_required:
                return Expressions.always_false()
            return BoundPredicate(Operation.IS_NULL, bound_term, is_unary_predicate=True)
        elif self.op == Operation.NOT_NULL:
            if bound_term.ref.field.is_required:
                return Expressions.always_true()
            return BoundPredicate(Operation.NOT_NULL, bound_term, is_unary_predicate=True)
        elif self.op in [Operation.IS_NAN, Operation.NOT_NAN]:
            if not self.floating_type(bound_term.ref.type.type_id):
                raise ValidationException(f"{self.op} cannot be used with a non-floating column", ())
            return BoundPredicate(self.op, bound_term, is_unary_predicate=True)

        raise ValidationException(f"Operation must be in [IS_NULL, NOT_NULL, IS_NAN, NOT_NAN] was:{self.op}", ())

    def bind_in_operation(self, bound_term):
        from .expressions import Expressions

        def convert_literal(lit):
            converted = lit.to(bound_term)
            ValidationException.check(converted is not None,
                                      "Invalid Value for conversion to type %s: %s (%s)",
                                      (bound_term.type, lit, lit.__class__.__name__))
            return converted

        converted_literals = filter(lambda x: x != Literals.above_max() and x != Literals.below_min(),
                                    [convert_literal(lit) for lit in self.literals])
        if len(converted_literals) == 0:
            return Expressions.always_true() if Operation.NOT_IN else Expressions.always_false()
        literal_set = set(converted_literals)
        if len(literal_set) == 1:
            if self.op == Operation.IN:
                return BoundPredicate(Operation.EQ, bound_term, literal_set[0])
            elif self.op == Operation.NOT_IN:
                return BoundPredicate(Operation.NOT_EQ, bound_term, literal_set[0])
            else:
                raise ValidationException("Operation must be in or not in", ())

        return BoundPredicate(self.op, bound_term, literals=literal_set, is_set_predicate=True)

    def bind_literal_operation(self, bound_term):
        from .expressions import Expressions

        lit = self.lit.to(bound_term.type)
        ValidationException.check(lit is not None,
                                  "Invalid Value for conversion to type %s: %s (%s)",
                                  (bound_term.type, self.lit, self.lit.__class__.__name__))

        if lit == Literals.above_max():
            if self.op in [Operation.LT, Operation.LT_EQ, Operation.NOT_EQ]:
                return Expressions.always_true()
            elif self.op in [Operation.GT, Operation.GT_EQ, Operation.EQ]:
                return Expressions.always_false()
        elif lit == Literals.below_min():
            if self.op in [Operation.LT, Operation.LT_EQ, Operation.NOT_EQ]:
                return Expressions.always_false()
            elif self.op in [Operation.GT, Operation.GT_EQ, Operation.EQ]:
                return Expressions.always_true()

        return BoundPredicate(self.op, bound_term, lit=lit, is_literal_predicate=True)

    @staticmethod
    def floating_type(type_id: TypeID) -> bool:
        return type_id in [TypeID.FLOAT, TypeID.DOUBLE]
